import logging

import nextcord

from bot_base import BotBase
from bot_base.context import BotContext

try:
    import nextcord as discord
    from nextcord.ext import commands  # noqa
except ModuleNotFoundError:
    import discord
    from discord.ext import commands  # noqa

log = logging.getLogger(__name__)


class Internal(commands.Cog):
    def __init__(self, bot):
        self.bot: BotBase = bot

    @commands.Cog.listener()
    async def on_ready(self):
        log.info(f"{self.__class__.__name__}: Ready")

    @commands.group(invoke_without_command=True)
    @commands.has_permissions(administrator=True)
    async def blacklist(self, ctx: BotContext) -> None:
        """Top level blacklist interface"""
        await ctx.send_help(ctx.command)

    @blacklist.group(invoke_without_command=True)
    @commands.has_permissions(administrator=True)
    async def add(self, ctx: BotContext) -> None:
        """Add something to the blacklist"""
        await ctx.send_help(ctx.command)

    @add.command(name="person")
    @commands.has_permissions(administrator=True)
    async def add_person(
        self, ctx: BotContext, user: discord.Object, *, reason=None
    ) -> None:
        """Add someone to the blacklist"""
        await self.bot.blacklist.add_to_blacklist(
            user.id, reason=reason, is_guild_blacklist=False
        )
        await ctx.send_basic_embed(f"I have added <@{user.id}> to the blacklist.")

    @add.command(name="guild")
    @commands.has_permissions(administrator=True)
    async def add_guild(
        self, ctx: BotContext, guild: discord.Object, *, reason=None
    ) -> None:
        await self.bot.blacklist.add_to_blacklist(
            guild.id, reason=reason, is_guild_blacklist=True
        )
        await ctx.send_basic_embed(
            f"I have added the guild `{guild.id}` to the blacklist"
        )

    @blacklist.command()
    @commands.has_permissions(administrator=True)
    async def list(self, ctx: BotContext) -> None:
        """List all current blacklists"""
        if self.bot.blacklist.users:
            user_blacklists = "\n".join(f"`{u}`" for u in self.bot.blacklist.users)
        else:
            user_blacklists = "No user's blacklisted."

        if self.bot.blacklist.guilds:
            guild_blacklists = "\n".join(f"`{g}`" for g in self.bot.blacklist.guilds)
        else:
            guild_blacklists = "No guild's blacklisted."

        await ctx.send(
            embed=discord.Embed(
                title="Blacklists",
                description=f"Users:\n{user_blacklists}\n\nGuilds:\n{guild_blacklists}",
            )
        )

    @blacklist.group(invoke_without_command=True)
    @commands.has_permissions(administrator=True)
    async def remove(self, ctx: BotContext) -> None:
        """Remove something from the blacklist"""
        await ctx.send_help(ctx.command)

    @remove.command(name="person")
    @commands.has_permissions(administrator=True)
    async def remove_person(self, ctx: BotContext, user: discord.Object) -> None:
        """Remove a person from the blacklist.

        Does nothing if they weren't blacklisted.
        """
        await self.bot.blacklist.remove_from_blacklist(
            user.id, is_guild_blacklist=False
        )
        await ctx.send_basic_embed("I have completed that action for you.")

    @remove.command(name="guild")
    @commands.has_permissions(administrator=True)
    async def remove_guild(self, ctx: BotContext, guild: discord.Object) -> None:
        """Remove a guild from the blacklist.

        Does nothing if they weren't blacklisted.
        """
        await self.bot.blacklist.remove_from_blacklist(
            guild.id, is_guild_blacklist=True
        )
        await ctx.send_basic_embed("I have completed that action for you.")


def setup(bot):
    bot.add_cog(Internal(bot))



from motor.motor_asyncio import AsyncIOMotorClient

from bot_base.db.document import Document


class MongoManager:
    def __init__(self, connection_url, database_name=None):
        database_name = database_name or "production"

        self.__mongo = AsyncIOMotorClient(connection_url)
        self.db = self.__mongo[database_name]

    def typed_lookup(self, attr: str) -> Document:
        return getattr(self, attr)

    def __getattr__(self, item) -> Document:
        """
        Parameters
        ----------
        item : str
            Denotes the 'table' to return

        Returns
        -------
        Document
            A Document made for said item
        """
        doc: Document = Document(self.db, item)
        setattr(self, item, doc)

        return doc



import asyncio
from typing import Optional

try:
    import nextcord as discord
except ModuleNotFoundError:
    import discord

from . import channel


# noinspection PyUnresolvedReferences
class Meta:
    """
    Used to inject functionality into multiple
    class's and reduce code duplication
    """

    async def prompt(
        self,
        message: str,
        *,
        timeout=60.0,
        delete_after=True,
        author_id=None,
    ):
        """An interactive reaction confirmation dialog.
        Parameters
        -----------
        message: str
            The message to show along with the prompt.
        timeout: float
            How long to wait before returning.
        delete_after: bool
            Whether to delete the confirmation message after we're done.
        author_id: Optional[int]
            The member who should respond to the prompt. Defaults to the author of the
            Context's message.
        Returns
        --------
        Optional[bool]
            ``True`` if explicit confirm,
            ``False`` if explicit deny,
            ``None`` if deny due to timeout

        Taken from R.Danny
        """
        if not self.channel.permissions_for(self.me).add_reactions:
            raise RuntimeError("Bot does not have Add Reactions permission.")
        fmt = f"{message}\n\nReact with \N{WHITE HEAVY CHECK MARK} to confirm or \N{CROSS MARK} to deny."

        # Ensure we can gather author id
        try:
            author_id = (
                author_id or self.author.id or self.id
            )  # self.id for User/Member
        except AttributeError:
            if issubclass(type(self), channel.WrappedChannel):
                raise RuntimeError(
                    "Expected author_id when using prompt on a TextChannel"
                )

        msg = await self.send(fmt)
        confirm = None

        def check(payload):
            nonlocal confirm
            if payload.message_id != msg.id or payload.user_id != author_id:
                return False
            codepoint = str(payload.emoji)
            if codepoint == "\N{WHITE HEAVY CHECK MARK}":
                confirm = True
                return True
            elif codepoint == "\N{CROSS MARK}":
                confirm = False
                return True
            return False

        for emoji in ("\N{WHITE HEAVY CHECK MARK}", "\N{CROSS MARK}"):
            await msg.add_reaction(emoji)
        try:
            await self.bot.wait_for("raw_reaction_add", check=check, timeout=timeout)
        except asyncio.TimeoutError:
            confirm = None
        try:
            if delete_after:
                await msg.delete()
        finally:
            return confirm

    async def send_basic_embed(
        self,
        desc: str,
        *,
        color=None,
        target=None,
        contain_timestamp: bool = True,
        include_command_invoker: bool = True,
        **kwargs
    ) -> discord.Message:
        """Wraps a string to send formatted as an embed"""
        target = target or self.channel

        embed = discord.Embed(description=desc)

        if color:
            embed.colour = color

        if contain_timestamp and not isinstance(self, channel.WrappedChannel):
            embed.timestamp = self.message.created_at

        if include_command_invoker and not isinstance(self, channel.WrappedChannel):
            embed.set_footer(
                text=self.author.display_name, icon_url=self.author.avatar.url
            )

        return await target.send(embed=embed, **kwargs)

    async def get_input(
        self,
        title: str = None,
        description: str = None,
        *,
        timeout: int = 100,
        delete_after: bool = True,
        author_id=None,
    ) -> Optional[str]:
        if title and not description:
            embed = discord.Embed(
                title=title,
            )
        elif not title and description:
            embed = discord.Embed(
                description=description,
            )
        elif title and description:
            embed = discord.Embed(
                title=title,
                description=description,
            )
        else:
            raise RuntimeError("Expected atleast title or description")

        sent = await self.send(embed=embed)
        val = None

        try:
            author_id = (
                author_id or self.author.id or self.id
            )  # or self.id for User/Member
        except AttributeError:
            if issubclass(type(self), channel.WrappedChannel):
                raise RuntimeError(
                    "Expected author_id when using prompt on a TextChannel"
                )

        try:
            msg = await self.bot.wait_for(
                "message",
                timeout=timeout,
                check=lambda message: message.author.id == author_id,
            )
            if msg:
                val = msg.content
        except asyncio.TimeoutError:
            if delete_after:
                await sent.delete()

            return val

        try:
            if delete_after:
                await sent.delete()
                await msg.delete()
        finally:
            return val



try:
    from nextcord.ext import commands
except ModuleNotFoundError:
    from discord.ext import commands

from bot_base.wraps import Meta, WrappedChannel, WrappedMember, WrappedUser


class BotContext(commands.Context, Meta):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.message.channel = WrappedChannel(self.message.channel)

        if not self.guild:
            self.message.author = WrappedUser(self.message.author)
        else:
            self.message.author = WrappedMember(self.message.author)



from datetime import timedelta, datetime
from typing import Any, Dict

from bot_base.caches import Entry
from bot_base.caches.abc import Cache
from bot_base.exceptions import NonExistentEntry


class TimedCache(Cache):
    __slots__ = ("cache",)

    def __init__(self):
        self.cache: Dict[Any, Entry] = {}

    def __contains__(self, item: Any) -> bool:
        try:
            entry = self.cache[item]
            if entry.expiry_time < datetime.now():
                return False
        except KeyError:
            return False
        else:
            return True

    def add_entry(
        self, key: Any, value: Any, *, ttl: timedelta = None, override: bool = False
    ) -> None:
        self.cache[key] = Entry(value=value, expiry_time=(datetime.now() + ttl))

    def delete_entry(self, key: Any) -> None:
        try:
            self.cache.pop(key)
        except KeyError:
            pass

    def get_entry(self, key: Any) -> Any:
        if key not in self:
            raise NonExistentEntry

        return self.cache[key].value

    def force_clean(self) -> None:
        now = datetime.now()
        for k, v in self.cache.items():
            if v.expiry_time < now:
                self.delete_entry(k)



try:
    from nextcord import DiscordException
except ModuleNotFoundError:
    from discord import DiscordException


class PrefixNotFound(DiscordException):
    """A prefix for this guild was not found."""


class ExistingEntry(DiscordException):
    """An entry was already found in the cache with this key."""


class NonExistentEntry(DiscordException):
    """No entry found in the cache with this key."""



import re

time_regex = re.compile(r"(( ?(\d{1,5})(h|s|m|d))+)")
time_dict = {"h": 3600, "s": 1, "m": 60, "d": 86400}


async def time_convertor(argument):
    raise RuntimeWarning("Lol dont")
    args = argument.lower()
    matches = re.findall(time_regex, args)
    if not matches:
        return 0

    matches = matches[0][0].split(" ")
    matches = [filter(None, re.split(r'(\d+)', s)) for s in matches]
    time = 0
    for match in matches:
        key, value = match
        try:
            time += time_dict[value] * float(key)
        except KeyError:
            raise commands.BadArgument(
                f"{value} is an invalid time key! h|m|s|d are valid arguments"
            )
        except ValueError:
            raise commands.BadArgument(f"{key} is not a number!")
    return time


import datetime
import sys
import logging
import traceback
import warnings
from typing import Optional, List

import discord
import humanize

from bot_base.caches import TimedCache

try:
    from nextcord import DiscordException
    from nextcord.ext import commands
except ModuleNotFoundError:
    from discord import DiscordException
    from discord.ext import commands

from bot_base.blacklist import BlacklistManager
from bot_base.context import BotContext
from bot_base.db import MongoManager
from bot_base.exceptions import PrefixNotFound
from bot_base.wraps import (
    WrappedChannel,
    WrappedMember,
    WrappedChannelConvertor,
    WrappedMemberConvertor,
    WrappedUserConvertor,
    WrappedUser,
)

log = logging.getLogger(__name__)

try:
    from nextcord.ext.commands.converter import CONVERTER_MAPPING

    CONVERTER_MAPPING[discord.User] = WrappedUserConvertor
    CONVERTER_MAPPING[discord.Member] = WrappedMemberConvertor
    CONVERTER_MAPPING[discord.TextChannel] = WrappedChannelConvertor
except ModuleNotFoundError:
    warnings.warn(
        "You don't have overridden converters. "
        "Please open an issue and name the fork your using if you want them."
    )


class BotBase(commands.Bot):
    def __init__(self, *args, **kwargs) -> None:
        mongo_manager = kwargs.pop("mongo_manager", None)
        if not mongo_manager:
            self.db: MongoManager = MongoManager(
                kwargs.pop("mongo_url"), kwargs.pop("mongo_database_name", None)
            )
        else:
            self.db: MongoManager = mongo_manager

        self.blacklist: BlacklistManager = BlacklistManager(self.db)
        self._uptime: datetime.datetime = datetime.datetime.now(
            tz=datetime.timezone.utc
        )
        self.prefix_cache: TimedCache = TimedCache()

        self.DEFAULT_PREFIX: str = kwargs.get("command_prefix")  # type: ignore

        super().__init__(*args, **kwargs)

        if kwargs.pop("load_builtin_commands", None):
            self.load_extension("bot_base.cogs.internal")

    @property
    def uptime(self) -> datetime.datetime:
        return self._uptime

    def get_bot_uptime(self) -> str:
        return humanize.precisedelta(
            self.uptime - datetime.datetime.now(tz=datetime.timezone.utc)
        )

    async def on_ready(self) -> None:
        await self.blacklist.initialize()

    async def get_command_prefix(self, message: discord.Message) -> List[str]:
        try:
            prefix = await self.get_guild_prefix(guild_id=message.guild.id)

            if message.content.casefold().startswith(prefix.casefold()):
                # The prefix matches, now return the one the user used
                # such that dpy will dispatch the given command
                prefix_length = len(prefix)
                prefix = message.content[:prefix_length]

            return commands.when_mentioned_or(prefix)(self, message)

        except (AttributeError, PrefixNotFound):
            return commands.when_mentioned_or(self.DEFAULT_PREFIX)(self, message)

    async def get_guild_prefix(self, guild_id: Optional[int] = None) -> str:
        """
        Using a cached property fetch prefixes
        for a guild and return em.

        Parameters
        ----------
        guild_id : int
            The guild we want prefixes for

        Returns
        -------
        str
            The prefix

        Raises
        ------
        PrefixNotFound
            We failed to find and
            return a valid prefix
        """
        if guild_id in self.prefix_cache:
            return self.prefix_cache.get_entry(guild_id)

        prefix_data = self.db.config.find(guild_id)

        if not prefix_data:
            raise PrefixNotFound

        prefix: Optional[str] = prefix_data.get("prefix")
        if not prefix:
            raise PrefixNotFound

        self.prefix_cache.add_entry(guild_id, prefix, override=True)
        return prefix

    async def on_command_error(self, ctx: BotContext, error: DiscordException) -> None:
        error = getattr(error, "original", error)

        if isinstance(error, commands.NoPrivateMessage):
            await ctx.author.send("This command cannot be used in private messages.")
        elif isinstance(error, commands.DisabledCommand):
            await ctx.author.send("Sorry. This command is disabled and cannot be used.")
        elif isinstance(error, commands.CommandInvokeError):
            original = error.original
            if not isinstance(original, discord.HTTPException):
                print(f"In {ctx.command.qualified_name}:", file=sys.stderr)
                traceback.print_tb(original.__traceback__)
                print(f"{original.__class__.__name__}: {original}", file=sys.stderr)
        elif isinstance(error, commands.ArgumentParsingError):
            await ctx.send(error)
        elif isinstance(error, commands.PrivateMessageOnly):
            await ctx.send("This command can only be used in dm's.")
        elif isinstance(error, commands.NoPrivateMessage):
            await ctx.send("This command can only be used in Guilds.")

        if not isinstance(error, commands.CommandNotFound):
            if await self.db.command_usage.find(ctx.command.qualified_name) is None:
                await self.db.command_usage.upsert(
                    {
                        "_id": ctx.command.qualified_name,
                        "usage_count": 0,
                        "failure_count": 1,
                    }
                )
            else:
                await self.db.command_usage.increment(
                    ctx.command.qualified_name, 1, "failure_count"
                )

            log.debug(f"Command failed: `{ctx.command.qualified_name}`")
        raise error

    async def on_command_completion(self, ctx: BotContext) -> None:
        if ctx.command.qualified_name == "logout":
            return

        if await self.db.command_usage.find(ctx.command.qualified_name) is None:
            await self.db.command_usage.upsert(
                {
                    "_id": ctx.command.qualified_name,
                    "usage_count": 1,
                    "failure_count": 0,
                }
            )
        else:
            await self.db.command_usage.increment(
                ctx.command.qualified_name, 1, "usage_count"
            )
        log.debug(f"Command executed: `{ctx.command.qualified_name}`")

    async def on_guild_join(self, guild: discord.Guild) -> None:
        if guild.id in self.blacklist.guilds:
            log.info("Leaving blacklisted Guild(id=%s)", guild.id)
            await guild.leave()

    async def process_commands(self, message: discord.Message) -> None:
        ctx = await self.get_context(message, cls=BotContext)

        if ctx.author.id in self.blacklist.users:
            log.debug(f"Ignoring blacklisted user: {ctx.author.id}")
            return

        if ctx.guild is not None and ctx.guild.id in self.blacklist.guilds:
            log.debug(f"Ignoring blacklisted guild: {ctx.guild.id}")
            return

        await self.invoke(ctx)

    async def on_message(self, message: discord.Message) -> None:
        if message.author.bot:
            log.debug("Ignoring a message from a bot.")
            return

        await self.process_commands(message)

    async def get_or_fetch_member(self, guild_id: int, member_id: int) -> WrappedMember:
        """Looks up a member in cache or fetches if not found."""
        guild = await self.get_or_fetch_guild(guild_id)
        member = guild.get_member(member_id)
        if member is not None:
            return WrappedMember(member)

        member = await guild.fetch_member(member_id)
        return WrappedMember(member)

    async def get_or_fetch_channel(self, channel_id: int) -> WrappedChannel:
        """Looks up a channel in cache or fetches if not found."""
        channel = self.get_channel(channel_id)
        if channel:
            return WrappedChannel(channel)

        channel = await self.fetch_channel(channel_id)
        return WrappedChannel(channel)

    async def get_or_fetch_guild(self, guild_id: int) -> discord.Guild:
        """Looks up a guild in cache or fetches if not found."""
        guild = self.get_guild(guild_id)
        if guild:
            return guild

        guild = await self.fetch_guild(guild_id)
        return guild

    async def get_or_fetch_user(self, user_id: int) -> WrappedUser:
        """Looks up a user in cache or fetches if not found."""
        user = self.get_user(user_id)
        if user:
            return WrappedUser(user)

        user = await self.fetch_user(user_id)
        return WrappedUser(user)



import functools
from typing import List, Dict, Optional, Union, Any, TypeVar, Type

from motor.motor_asyncio import AsyncIOMotorDatabase  # type: ignore
from pymongo.results import DeleteResult  # type: ignore

T = TypeVar("T")


def return_converted(func):
    @functools.wraps(func)
    async def wrapped(*args, **kwargs):
        data: Union[Dict, List[Dict]] = await func(*args, **kwargs)

        self: Document = args[0]
        if not self.converter:
            return data

        if not isinstance(data, list):
            return self.converter(**data)

        new_data = []
        for d in data:
            new_data.append(self.converter(**d))

        return new_data

    return wrapped


class Document:
    _version = 7

    def __init__(
        self, database: AsyncIOMotorDatabase, document_name: str, converter=None
    ):
        self._document_name: str = document_name
        self._database: AsyncIOMotorDatabase = database
        self.document = database[document_name]

        self.converter: Type[T] = converter

    # <-- Pointer Methods -->
    async def find(self, data_id: Any) -> Union[Dict[str, Any], Type[T]]:
        return await self.find_by_id(data_id)

    async def delete(self, data_id: Any) -> Optional[DeleteResult]:
        return await self.delete_by_id(data_id)

    async def update(self, data: Dict[str, Any], *args: Any, **kwargs: Any) -> None:
        await self.update_by_id(data, *args, **kwargs)

    # <-- Actual Methods -->
    @return_converted
    async def get_all(
        self, filter_dict: Optional[Dict[str, Any]] = None, *args: Any, **kwargs: Any
    ) -> List[Optional[Union[Dict[str, Any], Type[T]]]]:
        filter_dict = filter_dict or {}

        return await self.document.find(filter_dict, *args, **kwargs).to_list(None)  # type: ignore

    @return_converted
    async def find_by_id(self, data_id: Any) -> Union[Dict[str, Any], Type[T]]:
        return await self.document.find_one({"_id": data_id})  # type: ignore

    @return_converted
    async def find_by_custom(
        self, filter_dict: Dict[str, Any]
    ) -> Optional[Union[Dict[str, Any], Type[T]]]:
        self.__ensure_dict(filter_dict)

        return await self.document.find_one(filter_dict)  # type: ignore

    @return_converted
    async def find_many_by_custom(
        self, filter_dict: Dict[str, Any]
    ) -> List[Union[Dict[str, Any], Type[T]]]:
        self.__ensure_dict(filter_dict)

        return await self.document.find(filter_dict).to_list(None)  # type: ignore

    async def delete_by_id(self, data_id: Any) -> Optional[DeleteResult]:
        if await self.find_by_id(data_id) is None:
            return None

        return await self.document.delete_many({"_id": data_id})

    async def delete_by_custom(
        self, filter_dict: Dict[str, Any]
    ) -> Optional[Union[List[DeleteResult], DeleteResult]]:
        self.__ensure_dict(filter_dict)

        if await self.find_by_custom(filter_dict) is None:
            return None

        return await self.document.delete_many(filter_dict)  # type: ignore

    async def insert(self, data: Dict[str, Any]) -> None:
        self.__ensure_dict(data)

        await self.document.insert_one(data)

    async def upsert(
        self, data: Dict[str, Any], option: str = "set", *args: Any, **kwargs: Any
    ) -> None:
        await self.update_by_id(data, option, upsert=True, *args, **kwargs)

    async def update_by_id(
        self, data: Dict[str, Any], option: str = "set", *args: Any, **kwargs: Any
    ) -> None:
        self.__ensure_dict(data)
        self.__ensure_id(data)

        if await self.find_by_id(data["_id"]) is None:
            return await self.insert(data)

        data_id = data.pop("_id")
        await self.document.update_one(
            {"_id": data_id}, {f"${option}": data}, *args, **kwargs
        )

    async def upsert_custom(
        self,
        filter_dict: Dict[str, Any],
        update_data: Dict[str, Any],
        option: str = "set",
        *args: Any,
        **kwargs: Any,
    ) -> None:
        await self.update_by_custom(
            filter_dict, update_data, option, upsert=True, *args, **kwargs
        )

    async def update_by_custom(
        self,
        filter_dict: Dict[str, Any],
        update_data: Dict[str, Any],
        option: str = "set",
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self.__ensure_dict(filter_dict)
        self.__ensure_dict(update_data)

        if not bool(await self.find_by_custom(filter_dict)):
            # Insert
            return await self.insert({**filter_dict, **update_data})

        # Update
        await self.document.update_one(
            filter_dict, {f"${option}": update_data}, *args, **kwargs
        )

    async def unset(self, data: Dict[str, Any]) -> None:
        self.__ensure_dict(data)
        self.__ensure_id(data)

        if await self.find_by_id(data["_id"]) is None:
            return

        data_id = data.pop("_id")
        await self.document.update_one({"_id": data_id}, {"$unset": data})

    async def increment(
        self, data_id: Any, amount: Union[int, float], field: str
    ) -> None:
        if await self.find_by_id(data_id) is None:
            return

        await self.document.update_one({"_id": data_id}, {"$inc": {field: amount}})

    # <-- Private methods -->
    @staticmethod
    def __ensure_dict(data: Dict[str, Any]) -> None:
        assert isinstance(data, dict)

    @staticmethod
    def __ensure_id(data: Dict[str, Any]) -> None:
        assert "_id" in data

    @property
    def document_name(self) -> str:
        return self._document_name

    @property
    def database(self) -> AsyncIOMotorDatabase:
        return self._database



from typing import Union

try:
    from nextcord import abc
    from nextcord.ext import commands
except ModuleNotFoundError:
    from discord import abc
    from discord.ext import commands

from bot_base.wraps.meta import Meta


class WrappedChannel(Meta, abc.GuildChannel, abc.PrivateChannel):  # noqa
    """Wraps nextcord.TextChannel for ease of stuff"""

    def __init__(self, channel: Union[abc.GuildChannel, abc.PrivateChannel]):
        self.channel: Union[abc.GuildChannel, abc.PrivateChannel] = channel

    def __getattr__(self, item):
        """Anything not found within Meta should be returned from channel itself"""
        return getattr(self.channel, item)

    @property
    def __class__(self):
        return type(self.channel)

    def __eq__(self, other) -> bool:
        if not isinstance(other, (type(self.channel), WrappedChannel)):
            return False

        if isinstance(other, type(self.channel)):
            return other.id == self.channel.id

        return other.channel.id == self.channel.id


class WrappedChannelConvertor(commands.TextChannelConverter):
    """Return WrappedMember on :nextcord.Member"""

    async def convert(self, ctx, argument: str) -> WrappedChannel:
        channel: Union[abc.GuildChannel, abc.PrivateChannel] = await super().convert(
            ctx=ctx, argument=argument
        )
        return WrappedChannel(channel)



from bot_base.db import MongoManager


class BlacklistManager:
    def __init__(self, db: MongoManager):
        self.db = db

        self.users = set()
        self.guilds = set()

    def __contains__(self, item: int) -> bool:
        """
        Checks whether an id is contained within
        an internal blacklist or not
        """
        assert isinstance(item, int)

        in_users = item in self.users
        in_guilds = item in self.guilds

        return in_users or in_guilds

    async def initialize(self) -> None:
        """
        Called sometime on creation in order to
        populate the internal blacklist.
        """
        all_guild_entries = await self.db.guild_blacklist.get_all()
        for guild in all_guild_entries:
            self.guilds.add(guild["_id"])

        all_user_entries = await self.db.user_blacklist.get_all()
        for user in all_user_entries:
            self.users.add(user["_id"])

    async def add_to_blacklist(
        self, item: int, reason: str = "Unknown", is_guild_blacklist: bool = True
    ) -> None:
        """
        Add a given int to the internal blacklist
        as well as persist it within the db
        """
        assert isinstance(item, int)

        if is_guild_blacklist:
            self.guilds.add(item)
            await self.db.guild_blacklist.upsert({"_id": item, "reason": reason})

        else:
            self.users.add(item)
            await self.db.user_blacklist.upsert({"_id": item, "reason": reason})

    async def remove_from_blacklist(
        self, item: int, is_guild_blacklist: bool = True
    ) -> None:
        """
        Removes a given blacklist, ignoring if
        they weren't blacklisted
        """
        assert isinstance(item, int)

        if is_guild_blacklist:
            self.guilds.discard(item)
            await self.db.guild_blacklist.delete(item)

        else:
            self.users.discard(item)
            await self.db.user_blacklist.delete(item)



from .bot import BotBase
import logging
from collections import namedtuple

__version__ = "1.2.7"


logging.getLogger(__name__).addHandler(logging.NullHandler())
VersionInfo = namedtuple("VersionInfo", "major minor micro releaselevel serial")
version_info = VersionInfo(
    major=1, minor=2, micro=7, releaselevel="production", serial=0
)



from datetime import timedelta
from typing import runtime_checkable, Protocol, Any


@runtime_checkable
class Cache(Protocol):
    def add_entry(
        self, key: Any, value: Any, *, ttl: timedelta = None, override: bool = False
    ) -> None:
        """
        Adds an entry to the cache with an optional time to live.

        Parameters
        ----------
        key: Any
            The key to store this value under
        value: Any
            The value for the given key
        ttl: timedelta, optional
            How long this entry should be valid for.
            Defaults to forever
        override: bool, optional
            If True, overrides an entry if it already exists

        Raises
        ------
        ExistingEntry
        """
        raise NotImplementedError

    def delete_entry(self, key: Any) -> None:
        """
        Deletes an entry in the cache.

        Parameters
        ----------
        key: Any
            The entry to delete

        Notes
        -----
        If the entry doesnt exist this won't act
        differently to if it did exist.
        """
        raise NotImplementedError

    def __contains__(self, item: Any) -> bool:
        """
        Returns True if the item exists in the cache.
        """
        raise NotImplementedError

    def force_clean(self) -> None:
        """
        Iterates over the cache, removing outdated entries.

        Implemented since by default the cache only cleans
        on access. I.e its lazy
        """
        raise NotImplementedError

    def get_entry(self, key: Any) -> Any:
        """
        Parameters
        ----------
        key: Any
            The key to get an entry for

        Returns
        -------
        Any
            The value for this if

        Raises
        ------
        NonExistentEntry
            Either the cache doesn't contain
            the key, or the Entry timed out
        """
        raise NotImplementedError



from .channel import WrappedChannel, WrappedChannelConvertor
from .meta import Meta
from .member import WrappedMember, WrappedMemberConvertor
from .user import WrappedUser, WrappedUserConvertor

__all__ = (
    "WrappedChannel",
    "WrappedChannelConvertor",
    "Meta",
    "WrappedMember",
    "WrappedMemberConvertor",
    "WrappedUser",
    "WrappedUserConvertor",
)



from typing import Any

try:
    import nextcord as discord
    from nextcord.ext import commands
except ModuleNotFoundError:
    import discord
    from discord.ext import commands

from bot_base.wraps.meta import Meta


class WrappedMember(Meta, discord.Member):
    """Wraps discord.Member for ease of stuff"""

    def __init__(self, person: discord.Member) -> None:
        self.person: discord.Member = person

    def __getattr__(self, item):
        """Anything not found within Meta should be returned from author itself"""
        return getattr(self.person, item)

    @property
    def __class__(self):
        return type(self.person)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, (type(self.person), WrappedMember)):
            return False

        if isinstance(other, type(self.person)):
            return other.id == self.person.id

        return other.person.id == self.person.id


class WrappedMemberConvertor(commands.MemberConverter):
    """Return WrappedMember on :discord.Member"""

    async def convert(self, ctx, argument: str) -> WrappedMember:
        member: discord.Member = await super().convert(ctx=ctx, argument=argument)
        return WrappedMember(member)
